﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public class GameManager : MonoBehaviour {

	private static GameManager _instance;
	public static GameManager Instance {
		get
		{
			if (!_instance)
			{
				_instance = GameObject.FindObjectOfType<GameManager>();
				if (!_instance)
				{
					GameObject container = new GameObject();
					container.name = "GameManager";
					_instance = container.AddComponent<GameManager>();
				}
			}
			
			return _instance;
		}
	}

	private Root _currentRoot;
	public Root CurrentRoot {
		get { return _currentRoot; }
		set { _currentRoot = value; }
	}

	private List<Root> _roots;
	public List<Root> Roots {
		get { return _roots; }
		set { _roots = value; }
	}

	private int _numLives = Constants.MAX_NUM_LIVES;
	public int NumLives {
		get {return _numLives; }
		set { _numLives = value; }
	}

	private int currentRootIndex = 0;
	private int totalRootsSpawned = 0;

	private LevelManager levelManager;
	private UIManager uiManager;

	private bool initialized = false;

	public void LoseLife() {
		_numLives--;
		uiManager.LoseLife();

		if (_numLives <= 0)
			GameOver();
	}

	private void GameOver() {
		_roots.Clear();
		initialized = false;
		CancelInvoke("GameUpdate");
		uiManager.DisplayGameOver((int)(-1*LowestRootPosition().y));
	}

	public void ChangeDirection(Direction newDirection) {
		if (initialized && _currentRoot.CurrentDirection != newDirection)
			_currentRoot.CurrentDirection = newDirection;
	}

	public void ChangeCurrentRoot(Vector2 mousePosition) {
		if (!initialized)
			return;

		Collider2D clickedCollider = Physics2D.OverlapPoint(mousePosition, 1 << Constants.ROOT_LAYER);

		if (clickedCollider != null)
		{
			Root selectedRoot = clickedCollider.transform.parent.parent.GetComponent<Root>();

			if (selectedRoot != _currentRoot && _roots.Contains(selectedRoot))
			{
				_currentRoot.Deselect();
				_currentRoot = selectedRoot;
				_currentRoot.Select();

				for (int i = 0; i < _roots.Count; i++)
				{
					if (_roots[i] == _currentRoot)
					{
						currentRootIndex = i;
						break;
					}
				}
			}
		}
	}

	public void NextRoot() {
		if (!initialized)
			return;

		_currentRoot.Deselect();

		if (_roots.Count > 0)
		{
			currentRootIndex = (currentRootIndex + 1) % _roots.Count;
			_currentRoot = _roots[currentRootIndex];

			_currentRoot.Select ();
		}
	}

	// Use this for initialization
	void Start () {
		levelManager = LevelManager.Instance;
		uiManager = UIManager.Instance;

		_roots = new List<Root>();

		StartCoroutine(PlayOpeningCutScene());
	}

	private IEnumerator PlayOpeningCutScene() {
		animation.Play();

		while (animation.isPlaying)
			yield return 0;

		GameObject root = new GameObject("Root");
		_currentRoot = root.AddComponent<Root>();
		_currentRoot.Instantiate(root, new Vector3(0f, -0.5f, 0f), Direction.DOWN);
		_currentRoot.Select();
		
		_roots.Add(_currentRoot);
		totalRootsSpawned++;

		initialized = true;
		
		InvokeRepeating("GameUpdate", 0.01f, Constants.MOVEMENT_RATE);
	}

	void GameUpdate () {
		float depth = LowestRootPosition().y;

		uiManager.UpdateDepthLabel((int)(-1*depth));

		if (depth < DepthForNewRoot())
		{
			SplitRoot();
		}

		for (int i = 0; i < _roots.Count; i++)
		{
			_roots[i].Move ();
		}

		levelManager.MoveCamera(depth);
	}

	public void RemoveRoot(Root rootToRemove) {
		if (!initialized)
			return;

		_roots.Remove(rootToRemove);

		if (_roots.Count == 0)
		{
			while (_numLives > 0)
				LoseLife();
		}
		else
		{
			LoseLife();

			if (rootToRemove == _currentRoot)
				NextRoot();
		}
	}

	public Vector2 LowestRootPosition() {
		Vector2 lowestPosition = _currentRoot.Position;

		for (int i = 0; i < _roots.Count; i++)
		{
			if (_roots[i].Position.y < lowestPosition.y)
				lowestPosition = _roots[i].Position;
		}

		return lowestPosition;
	}

	private float DepthForNewRoot() {
		return -25f * totalRootsSpawned;
	}

	private void SplitRoot() {
		Root randomRoot = _roots[Random.Range (0, _roots.Count)];
		int randomDirection = Random.Range (0f, 1f) < 0.5 ? -1 : 1;
		float distance = randomDirection*(randomRoot.WidthPixels/(float)Constants.PIXELS_PER_UNIT)*5.0f;

		if (Physics2D.OverlapPoint(new Vector2(randomRoot.Position.x + distance, 
		                                       randomRoot.Position.y)) == null 
		    && Physics2D.OverlapPoint(new Vector2(randomRoot.Position.x + distance, 
		                                      randomRoot.Position.y - Mathf.Abs (distance))) == null
		    && Physics2D.OverlapPoint(new Vector2(randomRoot.Position.x + distance/2.0f, 
		                                      randomRoot.Position.y)) == null 
		    && Physics2D.OverlapPoint(new Vector2(randomRoot.Position.x + distance, 
		                                      randomRoot.Position.y - Mathf.Abs (distance/2.0f))) == null)
		{
			GameObject root = new GameObject("Root");
			Root newRoot = root.AddComponent<Root>();
			newRoot.Instantiate(root, randomRoot.Position, randomDirection < 0 ? Direction.LEFT : Direction.RIGHT);
			
			_roots.Add (newRoot);
			totalRootsSpawned++;
		}
		else if (Physics2D.OverlapPoint(new Vector2(randomRoot.Position.x - distance, 
		                                            randomRoot.Position.y)) == null
		         && Physics2D.OverlapPoint(new Vector2(randomRoot.Position.x - distance, 
		                                      randomRoot.Position.y - Mathf.Abs (distance))) == null
		         && Physics2D.OverlapPoint(new Vector2(randomRoot.Position.x - distance/2.0f, 
		                                      randomRoot.Position.y)) == null 
		         && Physics2D.OverlapPoint(new Vector2(randomRoot.Position.x + distance, 
		                                      randomRoot.Position.y - Mathf.Abs (distance/2.0f))) == null)
		{
			GameObject root = new GameObject("Root");
			Root newRoot = root.AddComponent<Root>();
			newRoot.Instantiate(root, randomRoot.Position, randomDirection < 0 ? Direction.RIGHT : Direction.LEFT);
			
			_roots.Add (newRoot);
			totalRootsSpawned++;
		}
	}
}
